import logging

from typing import TypedDict, Optional, Any, List, Union

from dateutil import parser

from datacaster.classes import CastDataClass

logger = logging.getLogger(__name__)


ChoiceValue = Union[str, int, float]
ChoiceList = List[ChoiceValue]


class ChoiceDict(TypedDict):
    value: ChoiceValue
    order: int


class UploadedFile(TypedDict):
    """
    - Contains data about a file uploaded using [_input_file_](../ppa.ui/#input_file)

    !!! info "All UploadedFile Keys"

        __data__: File contents.

        __filename__: Original filename.

        __size__: Size of the file in bytes.

        __type__: Content type (if the browser was able to detect it).
    """

    data: str
    filename: str
    size: int
    type: str


class PayloadKey(TypedDict):
    """
    - Describes a key to get from the task payload

    !!! info "All PayloadKey Keys"

        __name__: The name of the key in the payload.

        __default_value__: An optional default value to use if the key is missing from the payload.
    """

    name: str
    default_value: Optional[Any] = None


class TimeStamp(TypedDict):
    """
    - Contains submitted date picker information

    !!! info "All TimeStamp Keys"

        __day__: The selected day.

        __month__: The selected month.

        __year__: The selected year.

        __hour__: The selected hour.

        __minute__: The selected minute.

        __second__: The submitted second.

        __timezone__: The chosen timezone.

        __timestamp__: The supplied date, time, & UTC offset in ISO-8601 format (`yyyy-mm-ddThh:mm:ss+hh:mm`).
    """

    hour: int
    minute: int
    second: int
    day: int
    month: int
    year: int
    timezone: str
    timestamp: str


class Vault(TypedDict):
    """
    - Contains information about a Vault configured in PPA

    !!! info "All Vault Keys"

        __name__: The name of the Vault.

        __url__: The Vault address.

        __driver__: The type of Vault.

        __expires_at__: The date the Vault access token will expire (if applicable).

    !!! example

        ```YAML linenums="1"
        name: AWS Secrets Manager
        url: secretsmanager.eu-west-2.amazonaws.com
        driver: aws-secrets-manager
        expires_at: 09:33:03 03/02/2022
        ```
    """

    name: str
    url: str
    driver: str
    expires_at: str


class VaultData(CastDataClass):
    __annotations__ = Vault.__annotations__
    __class_config__ = {
        "cast_functions": {
            "fields": {
                "url": lambda x: x.replace("https://", "").split(":")[0],
                "expires_at": lambda x: parser.isoparse(x).strftime("%H:%M:%S %d/%m/%Y") if x else "",
            }
        },
        "always_cast": ["url", "expires_at"],
    }


class TextConfig(TypedDict):
    name: str
    title: str
    default: Optional[str] = ""
    required: Optional[bool] = True
    pattern: Optional[str] = ""
    pattern_error: Optional[str] = None


class BuiltinTextConfig(TypedDict):
    name: str
    title: str
    default: Optional[str] = ""
    pattern_error: Optional[str] = None
    required: Optional[bool] = True


class ChoiceConfig(TypedDict):
    name: str
    title: str
    options: Union[ChoiceList, List[ChoiceDict]]
    default: ChoiceValue = ""
    required: Optional[bool] = True


class ChecklistConfig(TypedDict):
    name: str
    title: str
    options: Union[ChoiceList, List[ChoiceDict]]
    minimum: Optional[int] = None
    maximum: Optional[int] = None
    defaults: List[ChoiceValue] = []
    required: Optional[bool] = True


class MultilineConfig(TypedDict):
    name: str
    title: str
    pattern: Optional[str] = ""
    pattern_error: Optional[str] = None
    default: Optional[str] = ""
    lines: Optional[int] = 15
    required: Optional[bool] = True


class TimestampOrDateConfig(TypedDict):
    name: str
    title: str


class TextField(TypedDict):
    text: TextConfig


class EmailField(TypedDict):
    email: BuiltinTextConfig


class Ipv4AddressField(TypedDict):
    ipv4_address: BuiltinTextConfig


class HostnameField(TypedDict):
    hostname: BuiltinTextConfig


class NumberField(TypedDict):
    number: BuiltinTextConfig


class FloatField(TypedDict):
    float: BuiltinTextConfig


class ChecklistField(TypedDict):
    checklist: ChecklistConfig


class MultilineField(TypedDict):
    multiline: MultilineConfig


class ChoiceField(TypedDict):
    choice: ChoiceConfig


class TimestampField(TypedDict):
    timestamp: TimestampOrDateConfig


class DateField(TypedDict):
    date: TimestampOrDateConfig


AnyField = Union[
    TextField,
    EmailField,
    Ipv4AddressField,
    HostnameField,
    NumberField,
    FloatField,
    ChecklistField,
    MultilineField,
    ChoiceField,
    TimestampField,
    DateField,
]
FieldsList = List[AnyField]


class BaseField:
    def __init__(self, field_config, field_type) -> None:
        self.field_config = field_config
        self.field_type = field_type
        self.formatted_config = {
            "name": self.field_config.get("name"),
            "display_name": self.field_config.get("title"),
            "type": self.field_type,
            "options": {"required": self.field_config.get("required", True)},
        }


class TextInput(BaseField):
    def __init__(self, field_config, field_type="text") -> None:
        print(field_type)
        super().__init__(field_config, field_type)
        self.default_patten = "^(?!\\s*$).+"
        self.pattern = self.field_config.get("pattern", self.default_patten)

    @property
    def pattern_error(self):
        if not (pattern_error := self.field_config.get("pattern_error")):
            if self.pattern != self.default_patten:
                return f"Invalid input (must match: {self.pattern})"
            else:
                return "Input must not be blank"
        else:
            return pattern_error

    def format(self):
        self.formatted_config["options"].update(
            {"default": self.field_config.get("default"), "pattern": self.pattern, "patternMessage": self.pattern_error}
        )

        return self.formatted_config


class BuiltInInput(BaseField):
    def __init__(self, field_type=None, field_config=None) -> None:
        if field_config is None:
            field_config = {}
        super().__init__(field_config, "text")
        self.built_in_patterns = {
            "email": {
                "pattern": r"^[A-Za-z0-9][A-Za-z0-9\.\-_\+]+[A-Za-z0-9\-]@[A-Za-z0-9]{1}[A-Za-z0-9\.\-]*\.[A-Za-z0-9\.\-\.]*[A-Za-z]{2,}$",
                "patternMessage": "Invalid email address",
            },
            "ipv4_address": {
                "pattern": r"^(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$",
                "patternMessage": "Invalid IPv4 address",
            },
            "hostname": {
                "pattern": r"^(?=.{0,253}$)(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$",
                "patternMessage": "Not a valid hostname",
            },
            "number": {"pattern": r"^[0-9]+$", "patternMessage": "Invalid number"},
            "float": {"pattern": r"[0-9]+\.[0-9]+", "patternMessage": "Invalid floating point number"},
        }
        self.field_type = field_type

    def format(self):
        self.formatted_config["options"].update(
            {
                "default": self.field_config.get("default"),
                "pattern": self.built_in_patterns.get(self.field_type).get("pattern"),
                "patternMessage": self.field_config.get("pattern_error")
                or self.built_in_patterns.get(self.field_type).get("patternMessage"),
            }
        )

        return self.formatted_config


class ChoiceInput(BaseField):
    def __init__(self, field_config) -> None:
        super().__init__(field_config, "choice")
        self.initial_options = self.field_config.get("options")
        self.name = self.field_config.get("name")
        self.sorted_options = _sort_options(self.initial_options, self.field_config.get("default")) if field_config else None

    def format(self):
        self.formatted_config["options"].update(
            {"options": self.sorted_options, "default_option": self.field_config.get("default")}
        )

        return self.formatted_config


class ChecklistInput(BaseField):
    def __init__(self, field_config) -> None:
        super().__init__(field_config, "checklist")
        self.initial_options = self.field_config.get("options")
        self.name = self.field_config.get("name")
        self.sorted_options = _sort_options(self.initial_options, self.field_config.get("defaults")) if field_config else None

    def format(self):
        self.formatted_config["options"].update(
            {"options": self.sorted_options, "default_options": self.field_config.get("defaults")}
        )

        return self.formatted_config


class MultilineInput(TextInput):
    def __init__(self, field_config) -> None:
        super().__init__(field_config, "multiline")

    def format(self):
        self.formatted_config["options"].update(
            {
                "rows": self.field_config.get("lines", 15),
                "rowsMax": 100,
                "default": self.field_config.get("default"),
                "pattern": self.pattern,
                "patternMessage": self.pattern_error,
            }
        )

        return self.formatted_config


class TimestampInput(BaseField):
    def __init__(self, field_config) -> None:
        super().__init__(field_config, "timestamp")

    def format(self):
        return self.formatted_config


class DateInput(BaseField):
    def __init__(self, field_config) -> None:
        super().__init__(field_config, "date")

    def format(self):
        return self.formatted_config


def _to_list(value) -> list:
    return [value] if not isinstance(value, list) else value


def _process_default_values(options_list, defaults) -> list:
    if defaults_to_add := list(filter(lambda v: all((v not in options_list, v not in {"", None})), _to_list(defaults))):
        list(map(lambda v: options_list.insert(0, v), defaults_to_add))
    return options_list


def _sort_options_list(options, defaults=None) -> list:
    if defaults is None:
        return options
    return _process_default_values(options, defaults)


def _sort_options_dict(options, defaults=None) -> list:
    option_values = [option["value"] for option in sorted(options, key=lambda d: d["order"])]
    if defaults is None:
        return option_values
    return _process_default_values(option_values, defaults)


def _sort_options(options: Union[list, dict], defaults=None) -> list:
    if any((isinstance(field, dict) for field in options)):
        return _sort_options_dict(options, defaults)
    return _sort_options_list(options, defaults)
