"""
User Interface

This module contains actions for interacting with PPA Task Operators.

!!! question "PPA Task Interface"

    See [Task Interface](../task_interface) for more information on inputs & outputs in PPA.
"""

import re
import uuid
import logging

from typing import Optional, List, Dict, Union

import validators

from dateutil import parser

from ppa import ui

from .records import (
    UploadedFile,
    TimeStamp,
    FieldsList,
    TextInput,
    ChoiceInput,
    ChecklistInput,
    MultilineInput,
    TimestampInput,
    DateInput,
    BuiltInInput,
    ChoiceList,
    ChoiceDict,
    ChoiceValue,
    _sort_options,
)
from ._common import min_ppa_version, argument_minimum_versions

logger = logging.getLogger(__name__)


TableHeader = List[str]
TableRow = List[Union[str, int, bool, None]]
TableRows = List[TableRow]
TableTab = Dict[str, TableRows]

Color = Optional[Union[str, int]]
COLOR_MAP = {
    "yellow": "#FFEB3B",
    "orange": "#FB8C00",
    "blue": "#1565C0",
    "light_blue": "#4FC3F7",
    "dark_blue": "#283593",
    "green": "#27AE60",
    "light_green": "#82E0AA",
    "dark_green": "#145A32",
    "red": "#DC3831",
    "grey": "#757575",
    "purple": "#8E24AA",
    "black": "#000000",
    "pink": "#F06292",
}


def _validate_color(color: Optional[str] = None) -> Optional[str]:
    if color is None:
        return color

    if isinstance(color, int):
        color = str(color)

    # Tries to find the hex code associated with a color name...
    if hex := COLOR_MAP.get(color):
        return hex

    # ...then falls back to validating it as a hex code (both are supported)
    try:
        # Remove leading hash before testing the hex code
        int(color[1:] if color.startswith("#") else color, 16)
        return color if color.startswith("#") else f"#{color}"
    except ValueError:
        message = f"Supplied color '{color}' is not the name of a supported color or a hexadecimal code."
        ui.output_error(message)
        ui.output_error(f"Valid color names: {', '.join(COLOR_MAP.keys())}")
        raise ValueError(message)


@min_ppa_version("2.11.0")
@argument_minimum_versions(sensitive="3.0.0")
def input_form(text: str, fields: FieldsList, element_id: Optional[str] = None, sensitive: Optional[bool] = None) -> Dict:
    r"""Present a form with one or more fields for the user to complete & submit.

    Multiple fields types are supported, see [here](../form_inputs/#form-inputs) for more information.

    !!! question "Minimum Plugin Version: 7.0.0"

    ??? info "Input"

        - __text:__ the title of the input form

        - __fields:__ a list of [form fields](../form_inputs#field-types)

        - __sensitive:__ set to `true` to [protect sensitive data](../sensitive_data) (requires __PPA v3__ & __plugin version 8.0.0__)

    ??? success "Output"

        A dictionary containing the submitted form data.

        See [this example](../form_inputs#example) for more information on using form data.

    ??? example

        ```YAML linenums="1"
            - ppa.ui.input_form:
                text: New Joiner Form
                fields:

                  - text:
                      name: logon_name
                      title: Logon Name
                      pattern: >
                        [A-Za-z]+\.[A-Za-z]+
                      pattern_error: >
                        Logon name must be in [first name].[last name] format.

                  - email:
                      name: email_address
                      title: Email Address

                  - choice:
                      name: office
                      title: Office
                      options:
                        - London
                        - Paris
                        - New York
                      default: London

              save: joiner_data
        ```"""

    class_map = {
        "text": TextInput,
        "choice": ChoiceInput,
        "checklist": ChecklistInput,
        "multiline": MultilineInput,
        "timestamp": TimestampInput,
        "date": DateInput,
    }
    formatted_fields = []
    cast_fields = []
    choice_fields = []
    checklists = {}
    memory = None

    def _cast_value(value):
        try:
            if value.isdigit():
                return int(value)
            else:
                try:
                    return float(value)
                except ValueError:
                    return value
        except AttributeError:
            return value

    def _cast_value_list(items):
        return list(map(_cast_value, items))

    return_value_map = {
        "number": int,
        "float": float,
        "text": _cast_value,
        "choice": _cast_value,
        "checklist": _cast_value_list,
    }

    def _get_type_default_name(class_type):
        try:
            return next(iter([d for d in class_map[class_type]({}).format()["options"].keys() if "default" in d]))
        except StopIteration:
            return None

    field_names = list(map(lambda f: f[next(iter(f))]["name"], fields))

    if duplicate_field_names := set([name for name in field_names if field_names.count(name) > 1]):
        raise Exception(f"The following fields have been supplied more than once: '{', '.join(duplicate_field_names)}'.")

    for field in fields:
        for field_type, field_config in field.items():
            if field_type in BuiltInInput().built_in_patterns.keys():
                formatted_fields.append(BuiltInInput(field_type, field_config).format())
            else:
                try:
                    field_class = class_map[field_type](field_config)
                    formatted_fields.append(field_class.format())
                    if isinstance(field_class, ChoiceInput):
                        choice_fields.append(field_class)
                except KeyError:
                    raise Exception(f"Invalid field type '{field_type}', must be one of {', '.join(class_map.keys())}")
            if field_type in return_value_map:
                cast_fields.append({"type": field_type, "name": field_config["name"]})
            if field_type == "checklist":
                checklists[field_config["name"]] = {
                    "minimum": field_config.get("minimum"),
                    "maximum": field_config.get("maximum"),
                    "title": field_config["title"],
                }

    while True:
        if memory:
            for field in formatted_fields:
                if default_name := _get_type_default_name(field["type"]):
                    field["options"][default_name] = memory[field["name"]]

        ppa_response = ui.input_form(text, formatted_fields, element_id=element_id, sensitive=sensitive)
        for name, details in checklists.items():
            count = len(ppa_response[name])
            minimum = details["minimum"]
            maximum = details["maximum"]
            if minimum and count < minimum:
                ui.output_error(f"Please choose at least {minimum} record(s) for the {details['title']} field")
                continue
            if maximum and count > maximum:
                ui.output_error(f"Please choose no more than {maximum} record(s) for the {details['title']} field")
                continue
            checklists[name]["correct_no"] = True
        if all([values.get("correct_no") for values in checklists.values()]):
            break
        else:
            memory = ppa_response

    for field in cast_fields:
        name = field["name"]
        ppa_response[name] = return_value_map[field["type"]](ppa_response[name])
    return ppa_response


@argument_minimum_versions(sensitive="3.0.0", pattern_error="2.11.0")
def input_text(
    text: str,
    default: Optional[str] = None,
    pattern: str = ".*",
    pattern_error: Optional[str] = None,
    element_id: Optional[str] = None,
    required: bool = False,
    sensitive: Optional[bool] = None,
) -> str:
    r"""Present a text input field.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __text:__ the title of the input field

        - __default:__ an optional default value for the input field (defaults to empty)

        - __pattern:__ an optional regular expression to validate the supplied value (defaults )

        - __pattern_error:__ error to display if pattern validation fails (supported in PPA version 2.11.x & newer)

        - __required:__ set to `true` to require a value before submission (defaults to `false`)

        - __element_id:__ an optional ID for the interface element ([more info](../task_interface))

        - __sensitive:__ set to `true` to [protect sensitive data](../sensitive_data) (requires __PPA v3__ & __plugin version 8.0.0__)

    ??? success "Output"
        The string submitted by the task operator.

    ??? example
        Prompting for a username & saving it as a new variable called _username_.

        - The value is required, & must match the format _firstname.surname_

        - A folded string (the _>_ character) escapes any special pattern characters

        ```YAML linenums="1"
        - ppa.ui.input_text:
            text: Username
            required: true
            pattern: >
              [A-Za-z]+\.[A-Za-z]+
          save: username
        ```"""
    return ui.input_text(
        text,
        default=default,
        pattern=pattern,
        element_id=element_id,
        required=required,
        pattern_error=pattern_error,
        sensitive=sensitive,
    )


@argument_minimum_versions(sensitive="3.0.0", pattern_error="2.11.0")
def input_multiline_text(
    text: str,
    default: Optional[str] = None,
    pattern: str = ".*",
    pattern_error: Optional[str] = None,
    element_id: Optional[str] = None,
    required: bool = False,
    lines: int = 15,
    sensitive: Optional[bool] = None,
) -> str:
    r"""Present a multiline text input field.

    !!! question "Minimum Plugin Version: 3.4.0"

    ??? info "Input"

        - __text:__ the title of the input field

        - __default:__ a default value for the input field (defaults to empty)

        - __pattern:__ a regular expression to validate the supplied value

        - __pattern_error:__ error to display if pattern validation fails

        - __required:__ set to `true` to require a value before submission (defaults to `false`)

        - __element_id:__ an optional ID for the interface element ([more info](../task_interface))

        - __lines:__ Denotes how many lines the input element should have (defaults to 15)

        - __sensitive:__ set to `true` to [protect sensitive data](../sensitive_data) (requires __PPA v3__ & __plugin version 8.0.0__)

    ??? success "Output"
        The string submitted by the Task Operator. Any line breaks will be denoted by a `\n`.

    ??? example
        Prompting for a multiline comment & saving it as a new variable called _comment_.

        ```YAML linenums="1"
        - ppa.ui.input_multiline_text:
            text: Service Now Comment
            required: true
          save: comment
        ```"""
    return ui.input_text_area(
        text,
        default=default,
        pattern=pattern,
        element_id=element_id,
        required=required,
        rows=lines,
        max_rows=100,
        pattern_error=pattern_error,
        sensitive=sensitive,
    )


@argument_minimum_versions(sensitive="3.0.0")
def input_email(
    text: str, required: Optional[bool] = True, element_id: Optional[str] = None, sensitive: Optional[bool] = None
) -> Optional[str]:
    """Present a text input field used to supply an email address.

    The input will be validated, & the task operator will be re-prompted if validation fails.

    !!! question "Minimum Plugin Version: 5.3.0"

    ??? info "Input"

        - __text:__ the title of the input field

        - __required:__ set to `false` to make the field optional (defaults to `true`)

        - __element_id:__ an optional ID for the interface element ([more info](../task_interface))

        - __sensitive:__ set to `true` to [protect sensitive data](../sensitive_data) (requires __PPA v3__ & __plugin version 8.0.0__)

    ??? success "Output"

        The supplied email address, or _null_ if `required` is __false__ & no value was supplied.

    ??? example

        ```YAML linenums="1"
        - ppa.ui.input_email:
            text: Please supply your email address
          save: email_address
        ```"""
    while True:
        email_address = ui.input_text(text, element_id=element_id, required=required, sensitive=sensitive)
        if all([not required, not email_address]):
            return None
        if not validators.email(email_address):
            ui.output_error(f"Supplied email address '{email_address}' is not valid, please try again.")
            continue
        break
    return email_address


@argument_minimum_versions(sensitive="3.0.0")
def input_number(
    text: str,
    minimum: Optional[int] = None,
    maximum: Optional[int] = None,
    required: Optional[bool] = True,
    element_id: Optional[str] = None,
    sensitive: Optional[bool] = None,
) -> Optional[int]:
    """Present an input field used to supply a number.

    The input will be validated, & the task operator will be re-prompted if validation fails.

    !!! question "Minimum Plugin Version: 5.8.0"

    ??? info "Input"

        - __text:__ the title of the input field

        - __minimum:__ optional lowest acceptable number

        - __maximum:__ optional highest acceptable number

        - __required:__ set to `false` to make the field optional (defaults to `true`)

        - __element_id:__ an optional ID for the interface element ([more info](../task_interface))

        - __sensitive:__ set to `true` to [protect sensitive data](../sensitive_data) (requires __PPA v3__ & __plugin version 8.0.0__)

    ??? success "Output"

        The supplied number, or _null_ if `required` is __false__ & no value was supplied.

    ??? example

        ```YAML linenums="1"
        - ppa.ui.input_number:
            text: Supply the number of days
          save: days
        ```"""
    while True:
        value = ui.input_text(text, element_id=element_id, required=required, sensitive=sensitive)
        if not value and required is False:
            return None
        try:
            number = int(value)
        except (ValueError, TypeError):
            ui.output_error(f"Supplied value '{value}' is not a number, please try again.")
            continue
        if minimum and number < minimum:
            ui.output_error(f"The number must be at least {minimum}, please try again.")
            continue
        if maximum and number > maximum:
            ui.output_error(f"The number must be no more than {maximum}, please try again.")
            continue
        return number


def input_ipv4(text: str, required: Optional[bool] = True, element_id: Optional[str] = None) -> Optional[str]:
    """Present a text input field used to supply an IPv4 address.

    The input will be validated, & the task operator will be re-prompted if validation fails.

    !!! question "Minimum Plugin Version: 5.3.0"

    ??? info "Input"

        - __text:__ the title of the input field

        - __required:__ set to `false` to make the field optional (defaults to `true`)

        - __element_id:__ an optional ID for the interface element ([more info](../task_interface))

    ??? success "Output"

        The supplied IPv4 address, or _null_ if `required` is __false__ & no value was supplied.

    ??? example

        ```YAML linenums="1"
        - ppa.ui.input_ipv4:
            text: IP Address
          save: ip_address
        ```"""
    while True:
        ip_address = ui.input_text(text, element_id=element_id, required=required)
        if all([not required, not ip_address]):
            return None
        if not _valid_ipv4(ip_address):
            ui.output_error(f"Supplied IPv4 address '{ip_address}' is not valid, please try again.")
            continue
        break
    return ip_address


def input_ipv4_cidr(text: str, required: Optional[bool] = True, element_id: Optional[str] = None) -> Optional[str]:
    """Present a text input field used to supply an IPv4 CIDR address.

    The input will be validated, & the task operator will be re-prompted if validation fails.

    !!! question "Minimum Plugin Version: 5.3.0"

    ??? info "Input"

        - __text:__ the title of the input field

        - __required:__ set to `false` to make the field optional (defaults to `true`)

        - __element_id:__ an optional ID for the interface element ([more info](../task_interface))

    ??? success "Output"

        The supplied IPv4 CIDR address, or _null_ if `required` is __false__ & no value was supplied.

    ??? example

        ```YAML linenums="1"
        - ppa.ui.input_ipv4_cidr:
            text: CIDR Address
          save: ip_address
        ```"""
    while True:
        cidr = ui.input_text(text, element_id=element_id, required=required)
        if all([not required, not cidr]):
            return None
        if not _valid_ipv4_cidr(cidr):
            ui.output_error(f"Supplied IPv4 CIDR address '{cidr}' is not valid, please try again.")
            continue
        break
    return cidr


def _valid_ipv4(ipv4_add: str) -> bool:
    if _valid_octets(ipv4_add):
        return True if validators.ipv4(ipv4_add) else False
    return False


def _valid_ipv4_cidr(ipv4_add: str) -> bool:
    if validators.ipv4_cidr(ipv4_add):
        ipv4, cidr = re.split("/", ipv4_add, 1)
        try:
            if _valid_octets(ipv4) and len(cidr) == len(str(int(cidr))):
                return True
        except ValueError:
            return False
    return False


def _valid_octets(ipv4: str) -> bool:
    return all([len(octet) <= 3 and len(octet) == len(str(int(octet))) for octet in ipv4.split(".")])


def input_hostname(text: str, required: Optional[bool] = True, element_id: Optional[str] = None) -> Optional[str]:
    """Present a text input field used to supply a hostname.

    The hostname format will be validated, & the task operator will be re-prompted if validation fails.

    !!! question "Minimum Plugin Version: 5.3.0"

    ??? info "Input"

        - __text:__ the title of the input field

        - __required:__ set to `false` to make the field optional (defaults to `true`)

        - __element_id:__ an optional ID for the interface element ([more info](../task_interface))

    ??? success "Output"

        The supplied hostname, or _null_ if `required` is __false__ & no value was supplied.

    ??? example

        ```YAML linenums="1"
        - ppa.ui.input_hostname:
            text: Hostname
          save: hostname
        ```"""
    # There doesn't seem to be a library that does this for us, so I've used the regex from here:
    # https://stackoverflow.com/a/3824105
    regex = re.compile(
        r"^([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]{0,61}[a-zA-Z0-9])"
        r"(\.([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]{0,61}[a-zA-Z0-9]))*$"
    )
    while True:
        hostname = ui.input_text(text, element_id=element_id, required=required)
        if all([not required, not hostname]):
            return None
        if not re.match(regex, hostname):
            ui.output_error(f"Supplied hostname '{hostname}' is not valid, please try again.")
            continue
        break
    return hostname


@argument_minimum_versions(sensitive="3.0.0")
def input_choice(
    text: str,
    options: Union[ChoiceList, List[ChoiceDict]],
    default: ChoiceValue = None,
    element_id: Optional[str] = None,
    sensitive: Optional[bool] = None,
) -> str:
    """Present a dropdown & wait for a choice.

    !!! question "Minimum Plugin Version: 6.0.0"

    ??? info "Input"

        - __text:__ the title of the dropdown list

        - __options:__ one or more values to populate the dropdown list (see the examples for more information)

        - __default:__ optional value to use as a default in the dropdown list (defaults to first value in **options**)

        - __element_id:__ an optional ID for the interface element ([more info](../task_interface))

        - __sensitive:__ set to `true` to [protect sensitive data](../sensitive_data) (requires __PPA v3__ & __plugin version 8.0.0__)

    ??? success "Output"

        The value chosen by the task operator.

    ??? example "Example 1: Simple"

        Prompting for a team & saving the choice as a new variable called _team_.

        ```YAML linenums="1"
        - ppa.ui.input_choice:
            text: Select a Team
            options:
              - Sales
              - Marketing
              - Engineering
          save: team
        ```

    ??? example "Example 2: Custom Ordering"

        !!! tip "Minimum Plugin Version: 7.0.0"

        ```YAML linenums="1"
          - ppa.ui.input_choice:
              text: Select a Quantity
              options:
                - value: Sales
                  order: 1
                - value: Marketing
                  order: 2
                - value: Engineering
                  order: 3
            save: quantity
        ```"""
    return ui.input_choice(
        text, options=_sort_options(options, default), default=default, element_id=element_id, sensitive=sensitive
    )


@argument_minimum_versions(sensitive="3.0.0")
def input_checklist(
    text: str,
    options: Union[ChoiceList, List[ChoiceDict]],
    defaults: List[Union[str, int]] = None,
    minimum: Optional[int] = None,
    maximum: Optional[int] = None,
    element_id: Optional[str] = None,
    sensitive: Optional[bool] = None,
) -> List[str]:
    """Present a list of checkboxes & wait for a selection to be made.

    !!! question "Minimum Plugin Version: 6.0.0"

    !!! question "Minimum PPA Version: 2.10.0"

    ??? info "Input"

        - __text:__ the title of the checklist

        - __options:__ one or more values to populate the dropdown list (see the examples for more information)

        - __defaults:__ optional list of values from **options** that will be checked by default

        - __element_id:__ an optional ID for the interface element ([more info](../task_interface))

        - __sensitive:__ set to `true` to [protect sensitive data](../sensitive_data) (requires __PPA v3__ & __plugin version 8.0.0__)

    ??? success "Output"

        A list of selected values (all will be strings).

    ??? example "Example 1: Simple"

        Prompting for a team & saving the selection as a new variable called _team_.

        ```YAML linenums="1"
        - ppa.ui.input_checklist:
            text: Select Groups
            options:
              - Domain Admin
              - PPA Admins
              - PAM Admins
              - PPA Access
              - Engineering
          save: team
        ```

    ??? example "Example 2: Custom Ordering"

        Prompting for groups to be selected & saving the selection as a new variable called _groups_

        - To avoid list values being automatically sorted, this example uses a dictionary as input

        - Each key is a dropdown item, & its value is its order in the list

        ```YAML linenums="1"
          - ppa.ui.input_checklist:
              text: Select Groups
              options:
                Domain Admin: 1
                PPA Admins: 2
                PPA Access: 3
                PAM Admins: 4
                Engineering: 5
            save: groups
        ```"""
    # Reverse the default options so if any are missing from the options provided, the order of the defaults is maintained.
    if defaults is not None:
        defaults.reverse()
    else:
        defaults = []

    if not element_id:
        element_id = str(uuid.uuid4())

    while True:
        choices = ui.input_checklist(
            text, options=_sort_options(options, defaults), defaults=defaults, element_id=element_id, sensitive=sensitive
        )
        count = len(choices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=f"{element_id}_error")
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=f"{element_id}_error")
            continue
        break
    return choices


def input_accept(text: str, element_id: Optional[str] = None) -> None:
    """Present a labelled button with text, & wait until it is clicked.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __text:__ the button label

        - __element_id:__ the ID of the element in the task interface ([more info](../task_interface))

    ??? success "Output"

        Nothing is returned from this action.

    ??? example
        ```YAML linenums="1"
        - ppa.ui.input_accept:
            text: Start
        ```"""
    ui.input_accept(text, element_id=element_id)


def input_password(text: str, meter: Optional[bool] = False, element_id: Optional[str] = None) -> str:
    """Present a password field.

    The password will be obfuscated in the interface & not stored in PPA.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __text:__ the title of the input field

        - __meter:__ set to `true` to show a password strength meter (defaults to `false`)

        - __element_id:__ an optional ID for the interface element ([more info](../task_interface))

    ??? success "Output"
        The password supplied by the task operator.

    ??? example
        ```YAML linenums="1"
        - ppa.ui.input_password:
            text: Enter a Password
            meter: true
          save: new_password
        ```"""
    return ui.input_password(text, meter=meter, element_id=element_id)


def input_password_confirm() -> str:
    """Present 'New Password' & 'Confirm Password' input fields.

    This action will loop until both supplied passwords match.

    The passwords are obfuscated in the interface & not stored in PPA.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        This action takes no inputs.

    ??? success "Output"

        The password supplied by the task operator.

    ??? example
        ```YAML linenums="1"
        - ppa.ui.input_password_confirm:
          save: new_password
        ```"""
    new_password_id, confirm_password_id, error_id = list(map(lambda _: str(uuid.uuid4()), range(0, 3)))

    while True:
        new_password = ui.input_password("New Password", meter=False, element_id=new_password_id)
        confirm_password = ui.input_password("Confirm Password", meter=False, element_id=confirm_password_id)
        if new_password == confirm_password:
            return new_password
        else:
            ui.output_error("Passwords must match.", element_id=error_id)


@argument_minimum_versions(sensitive="3.0.0")
def input_confirm(
    text: str,
    confirm: Optional[str] = "Yes",
    cancel: Optional[str] = "No",
    element_id: Optional[str] = None,
    sensitive: Optional[bool] = None,
) -> bool:
    """Present two labelled buttons to offer a choice, & wait until one is clicked.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __text:__ the text to display above the 2 buttons

        - __confirm:__ the label of the button that returns `true` (defaults to _Yes_)

        - __cancel:__ the label of the button that returns `false` (defaults to _No_)

        - __element_id:__ an optional ID for the task interface element ([more info](../task_interface))

        - __sensitive:__ set to `true` to [protect sensitive data](../sensitive_data) (requires __PPA v3__ & __plugin version 8.0.0__)

    ??? success "Output"
        A boolean (`true` or `false`).

    ??? example
        ```YAML linenums="1"
        - ppa.ui.input_confirm:
            text: Would you like to continue?
            confirm: Yes
            cancel: No
          save: confirm
        ```"""
    return ui.input_confirm(text, confirm, cancel, element_id=element_id, sensitive=sensitive)


@argument_minimum_versions(sensitive="3.0.0")
def input_table(
    text: str,
    header: TableHeader,
    rows: TableRows,
    minimum: Optional[int] = None,
    maximum: Optional[int] = None,
    element_id: Optional[str] = None,
    sensitive: Optional[bool] = None,
) -> List[int]:
    """Present a table & wait for a selection.

    !!! question "Minimum Plugin Version: 3.0.0"

    ??? info "Input"

        - __text:__ the table title

        - __header:__ a list of column names

        - __rows:__ the table contents as a list of lists, where each inner list is a row of values

        - __minimum:__ The minimum number of acceptable selections

        - __maximum:__ The maximum number of acceptable selections

        - __element_id:__ an optional ID for the interface element ([more info](../task_interface))

        - __sensitive:__ set to `true` to [protect sensitive data](../sensitive_data) (requires __PPA v3__ & __plugin version 8.0.0__)

    ??? success "Output"

        A list of the selected table row indices (zero-indexed).

    ??? example

        Presenting a table of cities & countries & saving the selection as `selected_rows`.

        ```YAML linenums="1" hl_lines="13"
        - ppa.ui.input_table:
            title: Select Cities
            header:
              - City
              - Country
            rows:
              - - Paris
                - France
              - - London
                - England
              - - Cardiff
                - Wales
          save: selected_rows
        ```

        If the _first_ & _second_ rows were chosen `selected_rows` would be:

        ```YAML linenums="1"
          - 0
          - 1
        ```"""
    while True:
        indices = ui.input_table(text, header=header, data=rows, element_id=element_id, sensitive=sensitive)
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=f"{element_id}_error")
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=f"{element_id}_error")
            continue
        break
    return indices


def input_file(text: str, accept: Union[str, List[str]] = None, element_id: Optional[str] = None) -> UploadedFile:
    """Ask the Task Operator to upload a file to the PPA task container.

    The file contents can be used & inspected by the task.

    The file will be made available to be downloaded by the task operator.

    !!! question "Minimum Plugin Version: 1.2.0"

    ??? info "Input"

        - __text:__ text to display with the upload component

        - __accept:__ optional file extension or MIME type to accept

        - __element_id:__ the ID of the element in the task interface ([more info](../task_interface))

    ??? success "Output"

        A single [UploadedFile](../inputs_outputs/#uploadedfile) dictionary.

    ??? example
        ```YAML linenums="1"
        - ppa.ui.input_file:
            text: Please upload a users CSV file
            accept: .csv
          save: users_csv
        ```"""
    return UploadedFile(**ui.input_file(text, accept=accept, element_id=element_id)._asdict())


def input_sensitive_file(text: str, accept: Union[str, List[str]] = None, element_id: Optional[str] = None) -> UploadedFile:
    """Ask the Task Operator to upload a file to the PPA task container.

    The file contents can be used & inspected by the task.

    The file will not be made available to be downloaded by the task operator.

    !!! question "Minimum PPA Appliance Version: 2.6.3"

    !!! question "Minimum Plugin Version: 3.4.0"

    ??? info "Input"

        - __text:__ text to display with the upload component

        - __accept:__ optional file extension or MIME type to accept

        - __element_id:__ the ID of the element in the task interface ([more info](../task_interface))

    ??? success "Output"
        A single [UploadedFile](../inputs_outputs/#uploadedfile) dictionary.

    ??? example
        ```YAML linenums="1"
        - ppa.ui.input_sensitive_file:
            text: Please upload a certificate private key
            accept: .pem
          save: private_key
        ```"""
    return UploadedFile(**ui.input_sensitive_file(text, accept=accept, element_id=element_id)._asdict())


@min_ppa_version("2.11.0")
def input_date(text: str) -> str:
    """Present a date picker & wait until it is submitted.

    !!! question "Minimum Plugin Version: 6.1.0"

    !!! question "Minimum PPA Version: 2.11.0"

    ??? info "Input"

        - __text:__ the text to display above the date picker

    ??? success "Output"

        A String.

    ??? example
        ```YAML linenums="1"
        - ppa.ui.input_date:
            text: Select a Start Date
          save: date
        ```"""

    return ui.input_date(text)


@min_ppa_version("2.8.0")
def input_timestamp(text: str) -> TimeStamp:
    """Present a datetime picker & wait until it is submitted.

    !!! question "Minimum Plugin Version: 5.2.0"

    !!! question "Minimum PPA Version: 2.8.0"

    ??? info "Input"

        - __text:__ the text to display above the datetime picker

    ??? success "Output"

        A [TimeStamp](../inputs_outputs/#timestamp).

    ??? example
        ```YAML linenums="1"
        - ppa.ui.input_timestamp:
            text: Select a Start Time
          save: timestamp
        ```"""
    response = ui.input_timestamp(text)
    parsed_timestamp = parser.isoparse(response["timestamp"])

    return TimeStamp(
        timezone=response["timezone"],
        timestamp=response["timestamp"],
        **{key: getattr(parsed_timestamp, key) for key in ["hour", "minute", "second", "day", "month", "year"]},
    )


@argument_minimum_versions(color="3.0.0", sensitive="3.0.0")
def output_info(text: str, color: Color = None, sensitive: Optional[bool] = None, element_id: Optional[str] = None) -> None:
    r"""Present some text in the task interface.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __text:__ the text to display

        - __element_id:__ an optional ID for the interface element ([more info](../task_interface))

        - __color\*:__ optional color name or hex code to apply to the text (__see below for more__)

        - __sensitive:__ set to `true` to [protect sensitive data](../sensitive_data) (requires __PPA v3__ & __plugin version 8.0.0__)

        ??? info "Colors"

            !!! question "Minimum PPA Version: 3.0.0"

            !!! question "Minimum Plugin Version: 8.0.0"

            Either a color name or a hex code can be supplied as the __color__ input.

            Valid color names:

                - yellow
                - orange
                - blue
                - light_blue
                - dark_blue
                - green
                - light_green
                - dark_green
                - red
                - grey
                - purple
                - black
                - pink

    ??? success "Output"

        Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - ppa.ui.output_info:
            text: Welcome to PPA
        ```"""
    ui.output_info(text, color=_validate_color(color), sensitive=sensitive, element_id=element_id)


@argument_minimum_versions(sensitive="3.0.0", color="3.0.0")
def output_error(text: str, sensitive: Optional[bool] = None, color: Color = None, element_id: Optional[str] = None) -> None:
    r"""Present an error message in the task interface.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __text:__ the text to display

        - __element_id:__ an optional ID for the interface element ([more info](../task_interface))

        - __color\*:__ optional color name or hex code to apply to the text (__see below for more__)

        - __sensitive:__ set to `true` to [protect sensitive data](../sensitive_data) (requires __PPA v3__ & __plugin version 8.0.0__)

        ??? info "Colors"

            !!! question "Minimum PPA Version: 3.0.0"

            !!! question "Minimum Plugin Version: 8.0.0"

            Either a color name or a hex code can be supplied as the __color__ input.

            Valid color names:

                - yellow
                - orange
                - blue
                - light_blue
                - dark_blue
                - green
                - light_green
                - dark_green
                - red
                - grey
                - purple
                - black
                - pink

    ??? success "Output"

        Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - ppa.ui.output_error:
            text: Something went wrong
        ```"""
    ui.output_error(text, color=_validate_color(color), sensitive=sensitive, element_id=element_id)


@argument_minimum_versions(sensitive="3.0.0")
def output_markdown(doc: str, sensitive: Optional[bool] = None, element_id: Optional[str] = None) -> None:
    r"""Present a markdown document to display.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __doc:__ the markdown document to display

        - __element_id:__ an optional ID for the interface element ([more info](../task_interface))

        - __sensitive:__ set to `true` to [protect sensitive data](../sensitive_data) (requires __PPA v3__ & __plugin version 8.0.0__)

    ??? success "Output"

        Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - ppa.ui.output_markdown:
            doc: >
              # Markdown Header

              Document text
        ```"""
    ui.output_markdown(doc, sensitive=sensitive, element_id=element_id)


def output_progress(text: str, percent: Optional[int] = None, element_id: Optional[str] = None) -> None:
    """Present a progress bar with an indeterminate or % value.

    The value _null_ is deemed indeterminate & will display an animation.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __text:__ the title of the progress bar

        - __percent:__ an integer progress value or `null`

        - __element_id:__ an optional ID for the interface element ([more info](../task_interface))

    ??? tip "Progress Bar Element IDs"
        It is essential to supply an _element_id_ to any progress bar whose value will be set more than once.

    ??? success "Output"
        Nothing is outputted by this action.

    ??? example
        - Showing an indeterminate progress bar

        - waiting for a button click

        - Updating the progress bar to 100%

        ```YAML linenums="1"
        - ppa.ui.output_progress:
            text: Waiting to continue...
            percent: null
            element_id: action_progress
        - ppa.ui.input_accept:
            text: Continue
        - ppa.ui.output_progress:
            text: Continuing
            percent: 100
            element_id: action_progress
        ```"""
    ui.output_progress(text, percent=percent, element_id=element_id)


@argument_minimum_versions(sensitive="3.0.0")
def output_table(
    text: str, header: TableHeader, rows: TableRows, element_id: Optional[str] = None, sensitive: Optional[bool] = None
) -> None:
    """Display a table.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __text:__ the table title

        - __header:__ a list of column names

        - __rows:__ the table contents as a list of lists, where each inner list is a row of values

        - __element_id:__ an optional ID for the interface element ([more info](../task_interface))

        - __sensitive:__ set to `true` to [protect sensitive data](../sensitive_data) (requires __PPA v3__ & __plugin version 8.0.0__)

    ??? success "Output"
        Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - ppa.ui.output_table:
            text: Cities
            header:
              - Name
              - Country
            rows:
              - - Paris
                - France
              - - London
                - England
              - - Brussels
                - Belgium
        ```"""
    ui.output_table(text, header=header, data=rows, sensitive=sensitive, element_id=element_id)


@argument_minimum_versions(sensitive="3.0.0")
def output_tabbed_table(
    text: str, header: TableHeader, tabs: TableTab, element_id: Optional[str] = None, sensitive: Optional[bool] = None
) -> None:
    """Display a table with multiple tabs.

    !!! question "Minimum Plugin Version: 3.5.0"

    ??? info "Input"

        - __text:__ the table title

        - __header:__ a list of column names

        - __tabs:__ a dictionary containing tab names & their rows as a list of lists

        - __element_id:__ an optional ID for the interface element ([more info](../task_interface))

        - __sensitive:__ set to `true` to [protect sensitive data](../sensitive_data) (requires __PPA v3__ & __plugin version 8.0.0__)

    ??? success "Output"

        Nothing is outputted by this action.

    ??? example

        Presenting a table with 2 tabs, where each tab contains a table with 3 columns & 1 row.

        ```YAML linenums="1"
        - ppa.ui.output_tabbed_table:
            text: Users
            header:
              - Team
              - Role
              - Location
            tabs:
              John Smith:
                - - Engineering
                  - Developer
                  - London
              David Bennett:
                - - Engineering
                  - Software Tester
                  - London
        ```"""
    ui.output_table(text, header=header, data=tabs, sensitive=sensitive, element_id=element_id)


def output_file(
    text: str, data: str, filename: str, type: str = "application/octet-stream", element_id: Optional[str] = None,
) -> None:
    """Allow the Task Operator to download a file from the PPA task.

    !!! question "Minimum Plugin Version: 1.2.0"

    ??? info "Input"

        - __text:__ text to display with the upload component

        - __data:__ file contents

        - __filename:__ name of the file

        - __type:__ optional MIME type (defaults to `application/octet-stream`)

        - __element_id:__ the ID of the element in the task interface ([more info](../task_interface))

    ??? success "Output"
        Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - ppa.ui.output_file:
            text: Please download the CSV report
            filename: report.csv
          load:
            data: csv_report
        ```
    """
    return ui.output_file(text, data=data, filename=filename, type=type, element_id=element_id)
